// (c) 2006 Philipp Klaus Krause philipp@colecovision.eu

#include <iostream>
#include <fstream>
#include <stack>
#include <string>

#include "cvu_compression.h"

extern const uint8_t HUFFMAN_ROOT;
extern const uint8_t HUFFMAN_LS, HUFFMAN_BS, HUFFMAN_RS;
extern const struct cvu_huffman_node huffman_tree[255];

#define cvu_huffman_right_end(node) ((node) >= HUFFMAN_BS)
#define cvu_huffman_left_end(node) ((node) >= HUFFMAN_LS && (node) < HUFFMAN_RS)

uint8_t output_byte;
unsigned char output_bit;	// Position in output_byte;
unsigned long output_count;

void put_bit(bool bit, std::ostream &out)
{
	output_byte |= (bit << output_bit);	

	output_bit++;
	if(output_bit == 8)
	{
		out << int(output_byte) << ", ";
		output_byte = 0;
		output_bit = 0;
		output_count++;
	}
}

bool find_current(uint8_t node, int current, std::stack<bool> &path)
{
	if(cvu_huffman_left_end(node) && (int(huffman_tree[node].left) == current))
	{
		path.push(false);
		return(true);
	}
	else if(cvu_huffman_right_end(node) && (int(huffman_tree[node].right) == current))
	{
		path.push(true);
		return(true);
	}
	else if(!cvu_huffman_left_end(node) && find_current(huffman_tree[node].left, current, path))
	{
		path.push(false);
		return(true);
	}
	else if(!cvu_huffman_right_end(node) && find_current(huffman_tree[node].right, current, path))
	{
		path.push(true);
		return(true);
	}
	else
	{
		return(false);
	}
}

void huffman_encode(std::istream &in, std::ostream &out)
{
	int current;
	output_bit = 0;
	output_byte = 0;

	while((current = in.get()) != EOF)
	{
		std::stack<bool> path;
		if(!find_current(HUFFMAN_ROOT, current, path))
		{
			std::cerr << "Internal error. Path not found in huffman tree.\n";
			return;
		}
		if(path.empty())
		{
			std::cerr << "Internal error. Empty path stack. n";
			return;
		}
		while(!path.empty())
		{
			put_bit(path.top(), out);
			path.pop();
		}
	}
	
	out << int(output_byte);
	output_count++;
}

int main(int argc, char *argv[])
{
	if(argc != 2)
	{
		std::cerr << "ColecoVision huffman encoder\n";
		std::cerr << "Usage: huffman_encoder <input file>\n";
		return(-1);
	}
	
	std::ifstream in(argv[1], std::ios_base::in | std::ios_base::binary);
	std::ofstream out((std::string(argv[1]) + ".c").c_str(), std::ios_base::out | std::ios_base::trunc);
	out << "//Huffman-encoded data encoded by ColecoVision huffman encoder.\n";
	out << "#include <stdint.h>\n\n";
	out << "const uint8_t data[] = { ";
	output_count = 0;
	huffman_encode(in, out);
	std::cout << "Output size: " << output_count << "\n";
	out << "};\n";

	in.close();
	out.close();

	return(0);
}
